"""Routines for automatically running functions and commands through `shadho`.

Functions
---------
load_config
    Load configurations sent from the `shadho` master.
run_task
    Run the user-defined task and save the result to file.
"""
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import copy
import json
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle
import tarfile
from typing import Iterable, Mapping

# If running on a remote system, try to import from the clobbered module name.
# Otherwise, import from the standard module name.
try:
    from shadho_utils import ShadhoEncoder, ShadhoDecoder
except ImportError:
    from utils import ShadhoEncoder, ShadhoDecoder


DEFAULTS = {
    'global': {
        'wrapper': 'shadho_run_task.py',
        'output': 'out.tar.gz',
        'result_file': 'performance.json',
        'optimize': 'loss',
        'param_file': 'hyperparameters.json'
    }
}


def load_config(path='.shadhorc'):
    """Load the configuration for this task.

    In most cases, the ``path`` parameter should be omitted in favor of using
    the configuration provided by the `shadho` master. This will ensure that
    the local configuration matches the global configuration and reduce
    headaches.

    Parameters
    ----------
    path : str, optional
        Path to the configuration file. It is recommended to use the default
        value in most cases. This will load the configuration provided by the
        `shadho` master and ensure that the local configuration matches the
        global configuration.

    Returns
    -------
    cfg : dict
        The configuration to use with this task.
    """
    config = copy.deepcopy(DEFAULTS)

    if os.path.isfile(path):
        cfg = configparser.ConfigParser()
        with open(path, 'r') as f:
            try:
                cfg.read_file(f)
            except AttributeError:
                cfg.readfp(f)
        for section in cfg.sections():
            for option in cfg.options(section):
                try:
                    t = type(DEFAULTS[section][option])
                except KeyError:
                    continue
                if t is bool:
                    val = cfg.getboolean(section, option)
                elif t is int:
                    val = cfg.getint(section, option)
                elif t is float:
                    val = cfg.getfloat(section, option)
                else:
                    val = cfg.get(section, option)

                config[section][option] = val

    return config


def run(task, cfgpath='.shadhorc'):
    """Run a function with arguments pulled from a JSON source.

    For simple use cases, SHADHO should be able to run without user-defined
    bash scripts, JSON parsing, etc. Users will expect this to be automated in
    the standard case.

    Parameters
    ----------
    task : callable
        The function/callable object to run.
    cfgpath : str, optional
        The path to the `shadho` configuration file for this task.
    """
    # Load the configuration
    cfg = load_config(path=cfgpath)

    # Load hyperparameters to test
    spec = {}
    if os.path.isfile(cfg['global']['param_file']):
        with open(cfg['global']['param_file'], 'r') as f:
            spec = json.load(f, cls=ShadhoDecoder)

    # Run the task and save the results in a format `shahdo` recognizes.
    result = task(spec)
    if isinstance(result, float) or isinstance(result, Iterable) and all([isinstance(i, float) for i in result]):
        result = {cfg['global']['optimize']: result}
    elif not isinstance(result, Mapping):
        raise ValueError(
            'Training result mush be a float, sequence of floats, ' +
            f'or dictionary of JSON-compatible results. Got {type(result)}')

    # Dump the results to file.
    with open(cfg['global']['result_file'], 'w') as f:
        json.dump(result, f, cls=ShadhoEncoder)

    # Compress into the expected output file.
    ext = os.path.splitext(cfg['global']['output'])
    mode = ':'.join(['w', ext]) if ext in ['gz', 'bz2'] else 'w'
    with tarfile.open(cfg['global']['output'], mode) as f:
        f.add(cfg['global']['result_file'])



"""Base class for SHADHO task managers.

Classes
-------
Manager
    Base class for task managers.
"""


class Manager():
    """Base class for task managers.

    Notes
    -----
    To implement a new local or distributed manager, create an __init__
    method and override all 
    """
    def __init__(self):
        pass

    def add_task(self, tag, cmd, params):
        """Add a task for the manger to run.

        Override this with all of the arguments necessary to run the task.

        Parameters
        ----------
        cmd
            Command or function to run in the task.
        tag : str
            Task id generated by SHADHO.
        params : dict
            Generarted hyperparameters to send to the task.
        """
        raise NotImplementedError

    def can_accept_tasks(self):
        """
        """
        raise NotImplementedError

    def handle_success(self):
        """
        """
        raise NotImplementedError

    def handle_failure(self):
        """
        """
        raise NotImplementedError

    def run_task(self):
        """
        """
        raise NotImplementedError


import json
import re

import numpy as np


class ShadhoEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (np.ndarray, np.generic)):
            return {
                '__data': obj.tolist(),
                '__dtype': str(obj.dtype),
            }
        else:
            return super(ShadhoEncoder, self).default(obj)


class ShadhoDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super(ShadhoDecoder, self).__init__(
            *args, object_hook=self.object_hook, **kwargs)

    def object_hook(self, obj):
        if isinstance(obj, dict) and sorted(obj.keys()) == ['__data', '__dtype']:
            if isinstance(obj['__data'], list):
                arr = np.array(obj['__data']).astype(obj['__dtype'])
            else:
                arr = np.dtype(obj['__dtype']).type(obj['__data'])
            return arr
        else:
            return obj
